# from .openquake import NrmlDocument, OpenquakeSimplePshaAdapter
from .psha_adapter_interface import (
    ConfigPshaAdapterInterface,
    GMCMPshaAdapterInterface,
    ModelPshaAdapterInterface,
    PshaAdapterInterface,
    SourcePshaAdapterInterface,
)
